import ssl
from typing import Optional

import pika
import pika.exceptions


class ConnectionError(Exception):
    """Error related to client → broker connection."""
    pass


class PCM:
    def __init__(self,
                 hostname: pika.SSLContext = 'localhost',
                 port: int = 5672,
                 virtual_host: str = '/',
                 username: str = 'guest',
                 password: str = 'guest',
                 tls_on: bool = False,
                 server_name_indication: bool = False) -> None:
        self._channel: Optional[pika.channel.Channel] = None
        self._connection: Optional[pika.connection.Connection] = None

        if tls_on:
            options = {'context': ssl.create_default_context()}
            if server_name_indication:
                server_hostname = hostname
                if isinstance(server_name_indication, str):
                    server_hostname = server_name_indication
                options['server_hostname'] = server_hostname
            ssl_options = pika.SSLOptions(**options)
        else:
            ssl_options = None

        try:
            self.conn_params = pika.ConnectionParameters(
                host=hostname,
                port=port,
                virtual_host=virtual_host,
                credentials=pika.credentials.PlainCredentials(username,
                                                              password),
                ssl_options=ssl_options)
        except Exception as e:
            raise ConnectionError(
                'Invalid connection parameters: {}'.format(e))

    def open(self) -> None:
        try:
            self._connection = pika.BlockingConnection(self.conn_params)
        except Exception as e:
            raise ConnectionError(
                'Unable to connect to broker: {} {}'.format(type(e), str(e)))
        self._channel = self._connection.channel()
        self._channel.confirm_delivery()

    def close(self) -> None:
        try:
            self._connection.close()
        except pika.exceptions.ConnectionClosed:
            pass

    def __enter__(self) -> PCM:
        if self._channel is None or not self._connection.is_open:
            self.open()
        return self

    def __exit__(self, exc_type: str, exc: str, trace: str) -> None:
        if self._connection is not None and self._connection.is_open:
            self.close()

    @property
    def channel(self) -> pika.channel.Channel:
        return self._channel

    def get_queue_length(self, queue_name: str) -> str:
        q_declared = self._channel.queue_declare(queue_name, passive=True)
        return q_declared.method.message_count

    def sleep(self, interval: float) -> None:
        self._connection.sleep(interval)

    def publish(self,
                exchange: str,
                routing_key: str,
                message: str,
                content_type: str = 'text/plain',
                delivery_mode: int = 2) -> bool:
        try:
            self._channel.basic_publish(exchange,
                                        routing_key,
                                        message,
                                        pika.BasicProperties(
                                            content_type=content_type,
                                            delivery_mode=delivery_mode))
        except pika.exceptions.NackError:
            return False
        else:
            return True
